/*
 * Copyright (C) 2018 Magnus Bergman <magnus.bergman@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <stdint.h>
#include <string.h>                     /* strcasecmp() */

#include "ansi_dec.h"
#include "cons_buf.h"
#include "debug.h"
#include "term_dyn.h"


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    Console *console;
    Terminal *terminal;
    AnsiDecoder *decoder;
    abydos_plugin_info_func_t info_func;
    abydos_plugin_progress_func_t progress_func;
    void *userdata;
};


typedef enum{
    S_ON,
    S_OFF,
    S_ICE
}ConsoleState;


static cairo_bool_t
has_blink(Console *cons){
    int size = cons_buf_width(cons) * cons_buf_height(cons);
    const unsigned short *p = cons_buf_buffer(cons);
    while(size>0){
        if(*p & 0x8000)
            return 1;
        --size;
        ++p;
    }
    return 0;
}


#include "vgafont16.h"


const char **
ansi_get_mime_types()
{
    static const char *mime_types[] = {
        "text/x-ansi",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
ansi_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
#if 0
    if(!strcasecmp(mime_type,"text/x-ansi")){
        /* OK */
    }else{
        free(h);
        return NULL;
    }
#endif
    h->info = info;
    h->console = cons_buf_new();
    cons_buf_set_size(h->console,80,1);
    h->terminal = term_dyn_new(&cons_buf_if,h->console);
    h->decoder = ansi_decoder_new(&term_dyn_if,h->terminal);
    return h;
}


static void
ansi_free(abydos_plugin_handle_t *h)
{
    ansi_decoder_free(h->decoder);
    term_dyn_free(h->terminal);
    cons_buf_free(h->console);
    free(h);
}



static void
ansi_progressive_begin(abydos_plugin_handle_t *h,abydos_plugin_info_func_t info_func,abydos_plugin_progress_func_t progress_func,void *userdata)
{
    h->info_func = info_func;
    h->progress_func = progress_func;
    h->userdata = userdata;
}


static int
ansi_progressive_feed(abydos_plugin_handle_t *h,const char *data,size_t len)
{
    while(len>0){
        int ch = (unsigned char)*data++;
        --len;
        if(ansi_decoder_write(h->decoder,ch))
            continue;
        if(26==ch)
            break;
        term_dyn_write(h->terminal,ch);
    }
    return 0;
}


static int
ansi_progressive_end(abydos_plugin_handle_t *h)
{
    h->info->width = cons_buf_width(h->console)*9;
    h->info->height = cons_buf_height(h->console)*16;
    h->info->pixel_ratio = 8.0/9;
    h->info->threadsafe = 1;
    if(has_blink(h->console))
        h->info->frame_count = 2;
    if(h->info_func)
        h->info_func(h->userdata);
    if(h->progress_func){
        cairo_rectangle_int_t r;
        r.x = 0;
        r.y = 0;
        r.width = h->info->width;
        r.height = h->info->height;
        h->progress_func(h->userdata,&r);
    }
    return 0;
}



static void
set_color(cairo_t *cr,int index){
    static const struct{
        double r;
        double g;
        double b;
    }ega[16] = {
        {    0,    0,    0}, /* black */
        {    0,    0,2.0/3}, /* dark blue */
        {    0,2.0/3,    0}, /* dark green */
        {    0,2.0/3,2.0/3}, /* dark cyan */
        {2.0/3,    0,    0}, /* dark red */
        {2.0/3,    0,2.0/3}, /* dark magenta */
        {2.0/3,1.0/3,    0}, /* brown */
        {2.0/3,2.0/3,2.0/3}, /* light gray */
        {1.0/3,1.0/3,1.0/3}, /* dark gray */
        {1.0/3,1.0/3,    1}, /* light blue */
        {1.0/3,    1,1.0/3}, /* light green */
        {1.0/3,    1,    1}, /* light cyan */
        {    1,1.0/3,1.0/3}, /* light red */
        {    1,1.0/3,    1}, /* light magenta */
        {    1,    1,1.0/3}, /* yellow */
        {    1,    1,    1}, /* white */
    };
    cairo_set_source_rgb(cr,ega[index].r,ega[index].g,ega[index].b);
}

static void
draw_character(cairo_t *cr,int x,int y,int ch)
{
    uint8_t *fontchar = &vga_rom_16[ch*16];
    int ninth = ch>=192 && ch<224;
    int i;
    int j;
    for(j=0;j<16;++j){
        int pixels = *fontchar++;
        for(i=0;i<8;++i){
            pixels <<= 1;
            if(pixels & 0x100)
                cairo_rectangle(cr,x+i,y+j,1,1);
        }
        if(ninth && (pixels & 0x100))
            cairo_rectangle(cr,x+8,y+j,1,1);
    }
}


static void
draw_screen(cairo_t *cr,Console *cons,ConsoleState state)
{
    const unsigned short *src = cons_buf_buffer(cons);
    int width = cons_buf_width(cons);
    int height = cons_buf_height(cons);
    int x;
    int y;
    cairo_save(cr);
    for(y=0;y<height;++y){
        for(x=0;x<width;++x){
            unsigned short cell = *src++;
            if(S_ICE==state){
                set_color(cr,(cell>>12)&15);
            }else{
                set_color(cr,(cell>>12)&7);
            }
            cairo_rectangle(cr,x*9,y*16,9,16);
            cairo_fill(cr);
            if(!(S_OFF==state && (cell&0x8000))){
                set_color(cr,(cell>>8)&15);
                draw_character(cr,x*9,y*16,cell&255);
                cairo_fill(cr);
            }
        }
    }
    cairo_restore(cr);
}


static void
ansi_render(abydos_plugin_handle_t *h, cairo_t *cr,int page,cairo_bool_t *layer_visibility,int variant,double time,int frame)
{
    draw_screen(cr,h->console,frame&1 ? S_OFF : S_ON);
}


static void
ansi_get_duration(abydos_plugin_handle_t *h,double *frame_duration)
{
    int i;
    /* The blink is on for 16 frames, then off for 16 frames.
       This emulates a frame rate of 60FPS. */
    for(i=0;i<h->info->frame_count;++i)
        frame_duration[i] = 16.0/60;
}


const abydos_plugin_interface_t *
Abydos0(){
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = ansi_get_mime_types,
        .new = ansi_new,
        .free = ansi_free,
        .progressive_begin = ansi_progressive_begin,
        .progressive_feed = ansi_progressive_feed,
        .progressive_end = ansi_progressive_end,
        .render = ansi_render,
        .get_duration = ansi_get_duration
    };
    DEBUG_p("begin");
    return &p;
}
