/*
 * Copyright (C) 2018 Magnus Bergman <magnus.bergman@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <magick/MagickCore.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* strcasecmp() */

#include "debug.h"


typedef struct{
    cairo_surface_t *surface;
    double duration;
}multi_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    const char *magick_type;
    enum{
        MT_NONE,
        MT_PAGE,
        MT_LAYER,
        MT_FRAME
    }multi_type;
    multi_t *multi;
    int multi_count;
};


static void
magick_done()
{
    MagickCoreTerminus();
}


static const char **
magick_get_mime_types()
{
    static const char *mime_types[] = {
        "image/gif",
        "image/jpeg",
        "image/tiff",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
magick_new(const char *mime_type,abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->multi_type = MT_NONE;
    if(!strcasecmp(mime_type,"image/gif")){
        h->magick_type = "GIF";
        h->multi_type = MT_FRAME;
    }else if(!strcasecmp(mime_type,"image/jpeg")){
        h->magick_type = "JPEG";
    }else if(!strcasecmp(mime_type,"image/tiff")){
        h->magick_type = "TIFF";
        h->multi_type = MT_PAGE;
    }else{
        free(h);
        return NULL;
    }
    h->info = info;
    h->multi = NULL;
    h->multi_count = 0;
    return h;
}


static void
magick_free(abydos_plugin_handle_t *h)
{
    if(h->multi){
        int i;
        for(i=0;i<h->multi_count;++i)
            cairo_surface_destroy(h->multi[i].surface);
        free(h->multi);
    }
    free(h);
}


static cairo_surface_t *
surface_from_image(Image *image,ExceptionInfo *exception)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,image->columns,image->rows);
    PixelPacket *src;
    uint32_t *dst;
    int rowskip;
    int x;
    int y;

#if MagickLibVersion >= 0x640
    src = GetAuthenticPixels(image,0,0,image->columns,image->rows,exception);
#else
    src = (PixelPacket*)AcquireImagePixels(image,0,0,image->columns,image->rows,exception);
#endif
    if(exception->severity){
        /* Can't get the pixels for the sub image, strange but not fatal. */
        DEBUG_p("failed to get sub image");
        ClearMagickException(exception);
        return surface;
    }

    dst = (uint32_t*)cairo_image_surface_get_data(surface);
    rowskip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t) - image->columns;;

    for(y=0;y<image->rows;++y){
        for(x=0;x<image->columns;++x){
            *dst++ = ((src->red>>(MAGICKCORE_QUANTUM_DEPTH-8))<<16)
                    |((src->green>>(MAGICKCORE_QUANTUM_DEPTH-8))<<8)
                    |((src->blue>>(MAGICKCORE_QUANTUM_DEPTH-8))<<0)
                    |((255-(src->opacity>>(MAGICKCORE_QUANTUM_DEPTH-8)))<<24);
            ++src;
        }
        dst += rowskip;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


static int
magick_create_from_data(abydos_plugin_handle_t *h,const char *data,size_t len)
{
    ExceptionInfo *exception = AcquireExceptionInfo();
    ImageInfo *info = CloneImageInfo(NULL);
    Image *image = NULL;
    int ret = -1;
    int i;

    strcpy(info->magick,h->magick_type);
    image = BlobToImage(info,data,len,exception);
    if(exception->severity){
        h->info->error = "Invalid image";
        goto out;
    }
    h->info->width = image->columns;
    h->info->height = image->rows;
    h->multi_count = GetImageListLength(image);
    DestroyImageInfo(info);

    switch(h->multi_type){
    case MT_NONE:
        DEBUG_p("multiple sub images was not expected for an image of type %s",h->magick_type);
        break;
    case MT_PAGE:
        h->info->page_count = h->multi_count;
        break;
    case MT_LAYER:
        h->info->layer_count = h->multi_count;
        break;
    case MT_FRAME:
        h->info->frame_count = h->multi_count;
        break;
    }

    h->multi = (multi_t*)malloc(h->multi_count * sizeof(multi_t));
    for(i=0;i<h->info->frame_count;++i){
        Image *subimage = GetImageFromList(image,i);
        h->multi[i].surface = surface_from_image(subimage,exception);
        if(MT_FRAME == h->multi_type && subimage->ticks_per_second > 0){
            h->multi[i].duration = (double)subimage->delay / subimage->ticks_per_second;
        }else{
            h->multi[i].duration = -1;
        }
    }
    ret = 0;
out:
    if(image)
        DestroyImage(image);
    if(exception)
        DestroyExceptionInfo(exception);
    return ret;
}


static cairo_surface_t *
magick_get_image_surface(abydos_plugin_handle_t *h,int page,cairo_bool_t *visible_layers,int prefered_variant,double time,int frame)
{
    int index;
    switch(h->multi_type){
    case MT_PAGE:
        index = page;
        break;
    case MT_FRAME:
        index = frame;
        break;
    default:
        index = 0;
        break;
    }
    return cairo_surface_reference(h->multi[index].surface);
}


static void
magick_get_duration(abydos_plugin_handle_t *h, double *duration)
{
    int i;
    for(i=0;i<h->info->frame_count;++i)
        duration[i] = h->multi[i].duration;
}


const abydos_plugin_interface_t *
Abydos0(){
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .done = magick_done,
        .get_mime_types = magick_get_mime_types,
        .new = magick_new,
        .free = magick_free,
        .create_from_data = magick_create_from_data,
        .get_image_surface = magick_get_image_surface,
        .get_duration = magick_get_duration
    };
    //if(!IsMagickInstantiated())
        MagickCoreGenesis(NULL,0);
    return &p;
}
